/** @file lorentzidx.h
 *
 *  Interface to GiNaC's Lorentz indices. */

/*
 *  GiNaC Copyright (C) 1999-2001 Johannes Gutenberg University Mainz, Germany
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __GINAC_LORENTZIDX_H__
#define __GINAC_LORENTZIDX_H__

#include <string>
#include <vector>
#include "idx.h"

namespace GiNaC {


/** Class of indices for Lorentz tensors, to tell them apart from other
 *  index families like color indices. The indices of this class also
 *  support the case of P-O-decomposed D-dimensional spacetime, where the
 *  parallel space is a Minkowski space with integral dimension P and the
 *  orthogonal space is a Euclidean space with (possibly complex) dimension
 *  D-P. The dimension D of spacetime is represented by the global symbol
 *  that is returned by the Dim() function. */
class lorentzidx : public idx
{
	GINAC_DECLARE_REGISTERED_CLASS(lorentzidx, idx)

	friend class simp_lor;
	friend class scalar_products;

// member functions

	// other constructors
public:
	explicit lorentzidx(bool cov, bool oonly=false, unsigned dimp=0);
	explicit lorentzidx(const std::string & n, bool cov=false,
	                    bool oonly=false, unsigned dimp=0);
	explicit lorentzidx(const char * n, bool cov=false,
	                    bool oonly=false, unsigned dimp=0);
	explicit lorentzidx(unsigned v, bool cov=false);

	// functions overriding virtual functions from bases classes
public:
	void printraw(std::ostream & os) const;
	void printtree(std::ostream & os, unsigned indent) const;
	void print(std::ostream & os, unsigned upper_precedence=0) const;
	bool info(unsigned inf) const;

	// new virtual functions which can be overridden by derived classes
	// none

	// non-virtual functions in this class
public:
	/** Check whether index only lives in orthogonal space. */
	bool is_orthogonal_only(void) const {return orthogonal_only;}

	/** Return dimension of parallel space. */
	unsigned get_dim_parallel_space(void) const {return dim_parallel_space;}

	lorentzidx create_anonymous_representative(void) const; 

	// member variables
protected:
	bool orthogonal_only;        /**< Symbolic index only lives in orthogonal space? */
	unsigned dim_parallel_space; /**< Dimension of parallel space */
};

// utility functions
inline const lorentzidx &ex_to_lorentzidx(const ex &e)
{
	return static_cast<const lorentzidx &>(*e.bp);
}

ex Dim(void);

} // namespace GiNaC

#endif // ndef __GINAC_LORENTZIDX_H__
